from unittest import mock

import pytest
from sqlalchemy import func, text
from sqlalchemy.orm import aliased, Query
from threedi_schema import constants, models, ThreediDatabase
from threedi_schema.beta_features import BETA_COLUMNS, BETA_VALUES

from threedi_modelchecker.checks.other import (
    BetaColumnsCheck,
    BetaValuesCheck,
    ChannelManholeLevelCheck,
    ConnectionNodesDistance,
    ConnectionNodesLength,
    CrossSectionLocationCheck,
    CrossSectionSameConfigurationCheck,
    ImperviousNodeInflowAreaCheck,
    LinestringLocationCheck,
    OpenChannelsWithNestedNewton,
    PerviousNodeInflowAreaCheck,
    PotentialBreachInterdistanceCheck,
    PotentialBreachStartEndCheck,
    PumpStorageTimestepCheck,
    SpatialIndexCheck,
)
from threedi_modelchecker.model_checks import ThreediModelChecker

from . import factories


def test_connection_nodes_length(session):
    factories.GlobalSettingsFactory(epsg_code=28992)
    factories.WeirFactory(
        connection_node_start=factories.ConnectionNodeFactory(
            the_geom="SRID=4326;POINT(-0.38222995634060702 -0.13872239147499893)"
        ),
        connection_node_end=factories.ConnectionNodeFactory(
            the_geom="SRID=4326;POINT(-0.3822292515698168 -0.1387223869163263)"
        ),
    )
    weir_too_short = factories.WeirFactory(
        connection_node_start=factories.ConnectionNodeFactory(
            the_geom="SRID=4326;POINT(-0.38222938832999598 -0.13872236685816669)"
        ),
        connection_node_end=factories.ConnectionNodeFactory(
            the_geom="SRID=4326;POINT(-0.38222930900909202 -0.13872236685816669)"
        ),
    )

    check_length = ConnectionNodesLength(
        column=models.Weir.id,
        start_node=models.Weir.connection_node_start,
        end_node=models.Weir.connection_node_end,
        min_distance=0.05,
    )

    errors = check_length.get_invalid(session)
    assert len(errors) == 1
    assert errors[0].id == weir_too_short.id


def test_connection_nodes_length_missing_start_node(session):
    factories.GlobalSettingsFactory(epsg_code=28992)
    factories.WeirFactory(
        connection_node_start_id=9999,
        connection_node_end=factories.ConnectionNodeFactory(
            the_geom="SRID=4326;POINT(-0.38222930900909202 -0.13872236685816669)"
        ),
    )

    check_length = ConnectionNodesLength(
        column=models.Weir.id,
        start_node=models.Weir.connection_node_start,
        end_node=models.Weir.connection_node_end,
        min_distance=0.05,
    )

    errors = check_length.get_invalid(session)
    assert len(errors) == 0


def test_connection_nodes_length_missing_end_node(session):
    if session.bind.name == "postgresql":
        pytest.skip("Postgres only accepts coords in epsg 4326")
    factories.GlobalSettingsFactory(epsg_code=28992)
    factories.WeirFactory(
        connection_node_start=factories.ConnectionNodeFactory(
            the_geom="SRID=4326;POINT(-0.38222930900909202 -0.13872236685816669)"
        ),
        connection_node_end_id=9999,
    )

    check_length = ConnectionNodesLength(
        column=models.Weir.id,
        start_node=models.Weir.connection_node_start,
        end_node=models.Weir.connection_node_end,
        min_distance=0.05,
    )

    errors = check_length.get_invalid(session)
    assert len(errors) == 0


def test_open_channels_with_nested_newton(session):
    factories.NumericalSettingsFactory(use_of_nested_newton=0)
    channel = factories.ChannelFactory(
        connection_node_start=factories.ConnectionNodeFactory(
            the_geom="SRID=4326;POINT(-71.064544 42.28787)"
        ),
        connection_node_end=factories.ConnectionNodeFactory(
            the_geom="SRID=4326;POINT(-71.0645 42.287)"
        ),
        the_geom="SRID=4326;LINESTRING(-71.064544 42.28787, -71.0645 42.287)",
    )
    open_definition = factories.CrossSectionDefinitionFactory(
        shape=constants.CrossSectionShape.TABULATED_TRAPEZIUM, width="1 0"
    )
    factories.CrossSectionLocationFactory(
        channel=channel,
        definition=open_definition,
        the_geom="SRID=4326;POINT(-71.0645 42.287)",
    )

    channel2 = factories.ChannelFactory(
        connection_node_start=factories.ConnectionNodeFactory(
            the_geom="SRID=4326;POINT(-71.064544 42.28787)"
        ),
        connection_node_end=factories.ConnectionNodeFactory(
            the_geom="SRID=4326;POINT(-71.0645 42.287)"
        ),
        the_geom="SRID=4326;LINESTRING(-71.064544 42.28787, -71.0645 42.287)",
    )
    open_definition_egg = factories.CrossSectionDefinitionFactory(
        shape=constants.CrossSectionShape.EGG,
    )
    factories.CrossSectionLocationFactory(
        channel=channel2,
        definition=open_definition_egg,
        the_geom="SRID=4326;POINT(-71.0645 42.287)",
    )

    check = OpenChannelsWithNestedNewton()

    errors = check.get_invalid(session)
    assert len(errors) == 2


channel_manhole_level_testdata = [
    ("start", -1, -3, -2, 0),
    ("start", -3, -1, -2, 1),
    ("end", -3, -1, -2, 0),
    ("end", -1, -3, -2, 1),
]


@pytest.mark.parametrize(
    "manhole_location,starting_reference_level,ending_reference_level,manhole_level,errors_number",
    channel_manhole_level_testdata,
)
def test_channel_manhole_level_check(
    session,
    manhole_location,
    starting_reference_level,
    ending_reference_level,
    manhole_level,
    errors_number,
):
    # using factories, create one minimal test case which passes, and one which fails
    # once that works, parametrise.
    # use nested factories for channel and connectionNode
    starting_coordinates = "4.718300 52.696686"
    ending_coordinates = "4.718255 52.696709"
    start_node = factories.ConnectionNodeFactory(
        the_geom=f"SRID=4326;POINT({starting_coordinates})"
    )
    end_node = factories.ConnectionNodeFactory(
        the_geom=f"SRID=4326;POINT({ending_coordinates})"
    )
    channel = factories.ChannelFactory(
        the_geom=f"SRID=4326;LINESTRING({starting_coordinates}, {ending_coordinates})",
        connection_node_start=start_node,
        connection_node_end=end_node,
    )
    # starting cross-section location
    factories.CrossSectionLocationFactory(
        the_geom="SRID=4326;POINT(4.718278 52.696697)",
        reference_level=starting_reference_level,
        channel=channel,
    )
    # ending cross-section location
    factories.CrossSectionLocationFactory(
        the_geom="SRID=4326;POINT(4.718264 52.696704)",
        reference_level=ending_reference_level,
        channel=channel,
    )
    # manhole
    factories.ManholeFactory(
        connection_node=end_node if manhole_location == "end" else start_node,
        bottom_level=manhole_level,
    )
    check = ChannelManholeLevelCheck(nodes_to_check=manhole_location)
    errors = check.get_invalid(session)
    assert len(errors) == errors_number


def test_node_distance(session):
    con1_too_close = factories.ConnectionNodeFactory(
        the_geom="SRID=4326;POINT(4.728282 52.64579283592512)"
    )
    con2_too_close = factories.ConnectionNodeFactory(
        the_geom="SRID=4326;POINT(4.72828 52.64579283592512)"
    )
    # Good distance
    factories.ConnectionNodeFactory(
        the_geom="SRID=4326;POINT(4.726838755789598 52.64514133594995)"
    )

    # sanity check to see the distances between the nodes
    node_a = aliased(models.ConnectionNode)
    node_b = aliased(models.ConnectionNode)
    distances_query = Query(
        func.ST_Distance(node_a.the_geom, node_b.the_geom, 1)
    ).filter(node_a.id != node_b.id)
    # Shows the distances between all 3 nodes: node 1 and 2 are too close
    distances_query.with_session(session).all()

    check = ConnectionNodesDistance(minimum_distance=10)
    invalid = check.get_invalid(session)
    assert len(invalid) == 2
    invalid_ids = [i.id for i in invalid]
    assert con1_too_close.id in invalid_ids
    assert con2_too_close.id in invalid_ids


@pytest.mark.parametrize(
    "channel_geom",
    [
        "LINESTRING(5.387204 52.155172, 5.387204 52.155262)",
        "LINESTRING(5.387218 52.155172, 5.387218 52.155262)",  # within tolerance
        "LINESTRING(5.387204 52.155262, 5.387204 52.155172)",  # reversed
        "LINESTRING(5.387218 52.155262, 5.387218 52.155172)",  # reversed, within tolerance
    ],
)
def test_channels_location_check(session, channel_geom):
    factories.ChannelFactory(
        connection_node_start=factories.ConnectionNodeFactory(
            the_geom="SRID=4326;POINT(5.387204 52.155172)"
        ),
        connection_node_end=factories.ConnectionNodeFactory(
            the_geom="SRID=4326;POINT(5.387204 52.155262)"
        ),
        the_geom=f"SRID=4326;{channel_geom}",
    )

    errors = LinestringLocationCheck(
        column=models.Channel.the_geom, max_distance=1.01
    ).get_invalid(session)
    assert len(errors) == 0


@pytest.mark.parametrize(
    "channel_geom",
    [
        "LINESTRING(5.387204 52.164151, 5.387204 52.155262)",  # startpoint is wrong
        "LINESTRING(5.387204 52.155172, 5.387204 52.164151)",  # endpoint is wrong
    ],
)
def test_channels_location_check_invalid(session, channel_geom):
    factories.ChannelFactory(
        connection_node_start=factories.ConnectionNodeFactory(
            the_geom="SRID=4326;POINT(5.387204 52.155172)"
        ),
        connection_node_end=factories.ConnectionNodeFactory(
            the_geom="SRID=4326;POINT(5.387204 52.155262)"
        ),
        the_geom=f"SRID=4326;{channel_geom}",
    )

    errors = LinestringLocationCheck(
        column=models.Channel.the_geom, max_distance=1.01
    ).get_invalid(session)
    assert len(errors) == 1


def test_cross_section_location(session):
    channel = factories.ChannelFactory(
        the_geom="SRID=4326;LINESTRING(5.387204 52.155172, 5.387204 52.155262)",
    )
    factories.CrossSectionLocationFactory(
        channel=channel, the_geom="SRID=4326;POINT(5.387204 52.155200)"
    )
    factories.CrossSectionLocationFactory(
        channel=channel, the_geom="SRID=4326;POINT(5.387218 52.155244)"
    )
    errors = CrossSectionLocationCheck(0.1).get_invalid(session)
    assert len(errors) == 1


@pytest.mark.parametrize(
    "shape, width, height, same_channels, ok",
    [
        # --- closed cross-sections ---
        # shapes 0, 2, 3 and 8 are always closed
        (0, "3", "4", True, False),
        *[(i, "3", None, True, False) for i in [2, 3, 8]],
        # shapes 5 and 6 are closed if the width at the highest increment (last number in the width string) is 0
        *[
            (
                i,
                "0 4.142 5.143 5.143 5.869 0",
                "0 0.174 0.348 0.522 0.696 0.87",
                True,
                False,
            )
            for i in [5, 6]
        ],
        # shape 7 is closed if the first and last (width, height) coordinates are the same
        (7, "2 4.142 5.143 5.143 5.869 2", "3 0.174 0.348 0.522 0.696 3", True, False),
        #
        # --- open cross-sections ---
        # shape 1 is always open
        (1, "3", "4", True, True),
        # shapes 5 and 6 are open if the width at the highest increment (last number in the width string) is > 0
        *[
            (
                i,
                "0 4.142 5.143 5.143 5.869 1",
                "0 0.174 0.348 0.522 0.696 0.87",
                True,
                True,
            )
            for i in [5, 6]
        ],
        # shape 7 is open if the first and last (width, height) coordinates are not the same
        # different width
        (7, "2 4.142 5.143 5.143 5.869 3", "4 0.174 0.348 0.522 0.696 4", True, True),
        # different height
        (7, "2 4.142 5.143 5.143 5.869 2", "3 0.174 0.348 0.522 0.696 4", True, True),
        # different height and width
        (7, "2 4.142 5.143 5.143 5.869 3", "4 0.174 0.348 0.522 0.696 5", True, True),
        #
        # Bad data, should silently fail, returning no invalid rows. The data is checked in other checks.
        (7, "foo", "bar", True, True),
        #
        # Check on different channels
        # this should fail if the cross-sections are on the same channel, but pass on different channels
        (0, "3", "4", False, True),
    ],
)
def test_cross_section_same_configuration(
    session, shape, width, height, same_channels, ok
):
    """
    This test checks two cross-sections on a channel against each other; they should both be open or both be closed.
    In this test, the first cross-section has been set to always be open.
    Therefore, the channel should be invalid when the second cross-section is closed, and valid when it is open.
    """
    first_channel = factories.ChannelFactory(
        the_geom="SRID=4326;LINESTRING(4.718301 52.696686, 4.718255 52.696709)",
    )
    second_channel = factories.ChannelFactory(
        the_geom="SRID=4326;LINESTRING(4.718301 52.696686, 4.718255 52.696709)",
    )
    # shape 1 is always open
    open_definition = factories.CrossSectionDefinitionFactory(
        id=1, shape=1, width="3", height="4"
    )
    factories.CrossSectionLocationFactory(
        channel=first_channel,
        the_geom="SRID=4326;POINT(4.718278 52.696697)",
        definition=open_definition,
    )
    testing_definition = factories.CrossSectionDefinitionFactory(
        id=2, width=width, height=height, shape=shape
    )
    # the second one is parametrised
    factories.CrossSectionLocationFactory(
        channel=first_channel if same_channels else second_channel,
        the_geom="SRID=4326;POINT(4.718265 52.696704)",
        definition=testing_definition,
    )
    errors = CrossSectionSameConfigurationCheck(models.Channel.id).get_invalid(session)
    assert len(errors) == (0 if ok else 1)


def test_spatial_index_ok(session):
    check = SpatialIndexCheck(models.ConnectionNode.the_geom)
    invalid = check.get_invalid(session)
    assert len(invalid) == 0


def test_spatial_index_disabled(empty_sqlite_v4):
    session = empty_sqlite_v4.get_session()
    session.execute(
        text("SELECT DisableSpatialIndex('v2_connection_nodes', 'the_geom')")
    )
    check = SpatialIndexCheck(models.ConnectionNode.the_geom)
    invalid = check.get_invalid(session)
    assert len(invalid) == 1


@pytest.mark.parametrize(
    "x,y,ok",
    [
        (-71.064544, 42.28787, True),  # at start
        (-71.0645, 42.287, True),  # at end
        (-71.06452, 42.2874, True),  # middle
        (-71.064544, 42.287869, False),  # close to start
        (-71.064499, 42.287001, False),  # close to end
    ],
)
def test_potential_breach_start_end(session, x, y, ok):
    # channel geom: LINESTRING (-71.064544 42.28787, -71.0645 42.287)
    factories.PotentialBreachFactory(
        the_geom=f"SRID=4326;LINESTRING({x} {y}, -71.064544 42.286)"
    )
    check = PotentialBreachStartEndCheck(
        models.PotentialBreach.the_geom, min_distance=1.0
    )
    invalid = check.get_invalid(session)
    if ok:
        assert len(invalid) == 0
    else:
        assert len(invalid) == 1


@pytest.mark.parametrize(
    "x,y,ok",
    [
        (-71.06452, 42.2874, True),  # exactly on other
        (-71.06452, 42.287401, False),  # too close to other
        (-71.0645, 42.287, True),  # far enough from other
    ],
)
def test_potential_breach_interdistance(session, x, y, ok):
    # channel geom: LINESTRING (-71.064544 42.28787, -71.0645 42.287)
    ref = factories.PotentialBreachFactory(
        the_geom="SRID=4326;LINESTRING(-71.06452 42.2874, -71.0646 42.286)"
    )
    factories.PotentialBreachFactory(
        the_geom=f"SRID=4326;LINESTRING({x} {y}, -71.064544 42.286)",
        channel=ref.channel,
    )
    check = PotentialBreachInterdistanceCheck(
        models.PotentialBreach.the_geom, min_distance=1.0
    )
    invalid = check.get_invalid(session)
    if ok:
        assert len(invalid) == 0
    else:
        assert len(invalid) == 1


def test_potential_breach_interdistance_other_channel(session):
    factories.PotentialBreachFactory(
        the_geom="SRID=4326;LINESTRING(-71.06452 42.2874, -71.0646 42.286)"
    )
    factories.PotentialBreachFactory(
        the_geom="SRID=4326;LINESTRING(-71.06452 42.287401, -71.064544 42.286)"
    )
    check = PotentialBreachInterdistanceCheck(
        models.PotentialBreach.the_geom, min_distance=1.0
    )
    invalid = check.get_invalid(session)
    assert len(invalid) == 0


@pytest.mark.parametrize(
    "storage_area,time_step,expected_result,capacity",
    [
        (0.64, 30, 1, 12.5),
        (600, 30, 0, 12.5),
        (None, 30, 0, 12.5),  # no storage --> open water --> no check
        (600, 30, 0, 0),
    ],
)
def test_pumpstation_storage_timestep(
    session, storage_area, time_step, expected_result, capacity
):
    connection_node = factories.ConnectionNodeFactory(storage_area=storage_area)
    factories.PumpstationFactory(
        connection_node_start=connection_node,
        start_level=-4,
        lower_stop_level=-4.78,
        capacity=capacity,
    )
    factories.GlobalSettingsFactory(sim_time_step=time_step)
    check = PumpStorageTimestepCheck(models.Pumpstation.capacity)
    invalid = check.get_invalid(session)
    assert len(invalid) == expected_result


@pytest.mark.parametrize(
    "value,expected_result",
    [
        (1000, 0),  # total area = 1000 + 9000 = 10000 <= 10000; no error
        (1001, 1),  # total area = 1001 + 9000 = 10001 > 10000; error
    ],
)
def test_impervious_connection_node_inflow_area(session, value, expected_result):
    connection_node = factories.ConnectionNodeFactory()
    first_impervious_surface = factories.ImperviousSurfaceFactory(area=9000)
    second_impervious_surface = factories.ImperviousSurfaceFactory(area=value)
    factories.ImperviousSurfaceMapFactory(
        impervious_surface=first_impervious_surface, connection_node=connection_node
    )
    factories.ImperviousSurfaceMapFactory(
        impervious_surface=second_impervious_surface, connection_node=connection_node
    )
    check = ImperviousNodeInflowAreaCheck()
    invalid = check.get_invalid(session)
    assert len(invalid) == expected_result


@pytest.mark.parametrize(
    "value,expected_result",
    [
        (1000, 0),  # total area = 1000 + 9000 = 10000 <= 10000; no error
        (1001, 1),  # total area = 1001 + 9000 = 10001 > 10000; error
    ],
)
def test_pervious_connection_node_inflow_area(session, value, expected_result):
    factories.ConnectionNodeFactory(id=1)
    factories.SurfaceFactory(id=1, area=9000)
    factories.SurfaceFactory(id=2, area=value)
    factories.SurfaceMapFactory(surface_id=1, connection_node_id=1)
    factories.SurfaceMapFactory(surface_id=2, connection_node_id=1)
    check = PerviousNodeInflowAreaCheck()
    invalid = check.get_invalid(session)
    assert len(invalid) == expected_result


@pytest.mark.parametrize(
    "value,expected_result",
    [
        (None, 0),  # column not set, valid result
        (5, 1),  # column set, invalid result
    ],
)
def test_beta_columns(session, value, expected_result):
    factories.GlobalSettingsFactory(vegetation_drag_settings_id=value)
    check = BetaColumnsCheck(models.GlobalSetting.vegetation_drag_settings_id)
    invalid = check.get_invalid(session)
    assert len(invalid) == expected_result


@pytest.mark.parametrize(
    "value,expected_result",
    [
        (
            constants.BoundaryType.RIEMANN,
            0,
        ),  # column not in beta columns, valid result
        (
            constants.BoundaryType.GROUNDWATERDISCHARGE,
            1,
        ),  # column in beta columns, invalid result
    ],
)
def test_beta_values(session, value, expected_result):
    beta_values = [
        constants.BoundaryType.GROUNDWATERLEVEL,
        constants.BoundaryType.GROUNDWATERDISCHARGE,
    ]
    factories.BoundaryConditions1DFactory(boundary_type=value)
    check = BetaValuesCheck(
        column=models.BoundaryCondition1D.boundary_type, values=beta_values
    )
    invalid = check.get_invalid(session)
    assert len(invalid) == expected_result


@pytest.mark.skipif(
    condition=(not BETA_COLUMNS and not BETA_VALUES),
    reason="requires beta features to be defined in threedi-schema to run",
)
@pytest.mark.parametrize(
    "allow_beta_features, no_checks_expected",
    [
        (False, False),
        (True, True),
    ],
)
def test_beta_features_in_server(threedi_db, allow_beta_features, no_checks_expected):
    with mock.patch.object(ThreediDatabase, "schema"):
        model_checker = ThreediModelChecker(
            threedi_db, allow_beta_features=allow_beta_features
        )
    model_beta_checks = [
        check
        for check in model_checker.config.checks
        if type(check) in [BetaColumnsCheck, BetaValuesCheck]
    ]
    if no_checks_expected:
        assert len(model_beta_checks) == 0
    else:
        assert len(model_beta_checks) > 0



import csv
from io import StringIO
from typing import NamedTuple

from threedi_modelchecker.checks.base import BaseCheck


# error handling export functions
def print_errors(errors):
    """Simply prints all errors to stdout

    :param errors: iterator of BaseModelError
    """
    for error in errors:
        print(format_check_results(*error))


def export_to_file(errors, file):
    """Write errors to a new file, separated with newlines.

    File cannot be an already existing file.

    :param errors: iterator of BaseModelError
    :param file:
    :return: None
    :raise FileExistsError: if the file already exists
    """
    with open(file, "w") as f:
        for error in errors:
            f.write(format_check_results(*error) + "\n")


def format_check_results(check: BaseCheck, invalid_row: NamedTuple):
    OUTPUT_FORMAT = "{level}{error_code:04d} (id={row_id:d}) {description!s}"
    return OUTPUT_FORMAT.format(
        level=check.level.name[:1],
        error_code=check.error_code,
        row_id=invalid_row.id,
        description=check.description(),
    )


# check overview export functions
def generate_rst_table(checks) -> str:
    "Generate an RST table to copy into the Sphinx docs with a list of checks"
    rst_table_string = ""
    header = (
        ".. list-table:: Executed checks\n"
        + "   :widths: 10 20 40\n"
        + "   :header-rows: 1\n\n"
        + "   * - Check number\n"
        + "     - Check level\n"
        + "     - Check message"
    )
    rst_table_string += header
    for check in checks:
        # pad error code with leading zeroes so it is always 4 numbers
        formatted_error_code = str(check.error_code).zfill(4)
        check_row = (
            "\n"
            + f"   * - {formatted_error_code}\n"
            + f"     - {check.level.name.capitalize()}\n"
            + f"     - {check.description()}"
        )
        rst_table_string += check_row
    return rst_table_string


def generate_csv_table(checks) -> str:
    "Generate an CSV table with a list of checks for use elsewhere"
    # a StringIO buffer is used so that the CSV can be printed to terminal as well as written to file
    output_buffer = StringIO()
    fieldnames = ["error_code", "level", "description"]
    writer = csv.DictWriter(
        output_buffer, fieldnames=fieldnames, quoting=csv.QUOTE_NONNUMERIC
    )

    writer.writeheader()

    for check in checks:
        writer.writerow(
            {
                "error_code": check.error_code,
                "level": check.level.name,
                "description": check.description(),
            }
        )

    return output_buffer.getvalue()



from dataclasses import dataclass
from typing import List, Literal, NamedTuple

from sqlalchemy import case, cast, distinct, func, REAL, select, text
from sqlalchemy.orm import aliased, Query, Session
from threedi_schema import constants, models

from .base import BaseCheck, CheckLevel
from .geo_query import distance, length, transform

# Use these to make checks only work on the first global settings entry:
first_setting = (
    Query(models.GlobalSetting.id)
    .order_by(models.GlobalSetting.id)
    .limit(1)
    .scalar_subquery()
)
first_setting_filter = models.GlobalSetting.id == first_setting


class CrossSectionLocationCheck(BaseCheck):
    """Check if cross section locations are within {max_distance} of their channel."""

    def __init__(self, max_distance, *args, **kwargs):
        super().__init__(column=models.CrossSectionLocation.the_geom, *args, **kwargs)
        self.max_distance = max_distance

    def get_invalid(self, session):
        # get all channels with more than 1 cross section location
        return (
            self.to_check(session)
            .join(models.Channel)
            .filter(
                distance(models.CrossSectionLocation.the_geom, models.Channel.the_geom)
                > self.max_distance
            )
            .all()
        )

    def description(self):
        return (
            f"v2_cross_section_location.the_geom is invalid: the cross-section location "
            f"should be located on the channel geometry (tolerance = {self.max_distance} m)"
        )


class CrossSectionSameConfigurationCheck(BaseCheck):
    """Check the cross-sections on the object are either all open or all closed."""

    def first_number_in_spaced_string(self, spaced_string):
        """return the first number in a space-separated string like '1 2 3'"""
        return cast(
            func.substr(
                spaced_string,
                1,
                func.instr(spaced_string, " ") - 1,
            ),
            REAL,
        )

    def last_number_in_spaced_string(self, spaced_string):
        """return the last number in a space-separated string like '1 2 3'"""
        return cast(
            func.replace(
                spaced_string,
                func.rtrim(
                    spaced_string,
                    func.replace(spaced_string, " ", ""),
                ),
                "",
            ),
            REAL,
        )

    def configuration_type(
        self, shape, first_width, last_width, first_height, last_height
    ):
        return case(
            (
                (
                    (shape.in_([0, 2, 3, 8]))
                    | (shape.in_([5, 6]) & (last_width == 0))
                    | (
                        (shape == 7)
                        & (first_width == last_width)
                        & (first_height == last_height)
                    )
                ),
                "closed",
            ),
            (
                (
                    (shape == 1)
                    | ((shape.in_([5, 6]) & (last_width > 0)))
                    | (
                        (shape == 7)
                        & ((first_width != last_width) | (first_height != last_height))
                    )
                ),
                "open",
            ),
            else_="open",
        )

    def get_invalid(self, session):
        # get all channels with more than 1 cross section location
        cross_sections = (
            select(
                models.CrossSectionLocation.id.label("cross_section_id"),
                models.CrossSectionLocation.channel_id,
                models.CrossSectionDefinition.shape,
                models.CrossSectionDefinition.width,
                models.CrossSectionDefinition.height,
                self.first_number_in_spaced_string(
                    models.CrossSectionDefinition.width
                ).label("first_width"),
                self.first_number_in_spaced_string(
                    models.CrossSectionDefinition.height
                ).label("first_height"),
                self.last_number_in_spaced_string(
                    models.CrossSectionDefinition.width
                ).label("last_width"),
                self.last_number_in_spaced_string(
                    models.CrossSectionDefinition.height
                ).label("last_height"),
            )
            .select_from(models.CrossSectionLocation)
            .join(
                models.CrossSectionDefinition,
                models.CrossSectionLocation.definition_id
                == models.CrossSectionDefinition.id,
            )
            .subquery()
        )
        cross_sections_with_configuration = select(
            cross_sections.c.cross_section_id,
            cross_sections.c.shape,
            cross_sections.c.last_width,
            cross_sections.c.channel_id,
            self.configuration_type(
                shape=cross_sections.c.shape,
                first_width=cross_sections.c.first_width,
                last_width=cross_sections.c.last_width,
                first_height=cross_sections.c.first_height,
                last_height=cross_sections.c.last_height,
            ).label("configuration"),
        ).subquery()
        filtered_cross_sections = (
            select(cross_sections_with_configuration)
            .group_by(cross_sections_with_configuration.c.channel_id)
            .having(
                func.count(distinct(cross_sections_with_configuration.c.configuration))
                > 1
            )
            .subquery()
        )

        def is_valid_series(input):
            try:
                [float(i) for i in input.split(" ")]
                return True
            except ValueError:
                return False

        def is_valid_value(input):
            if input in ["", None] or is_valid_series(input):
                return True
            else:
                return False

        all_cross_sections = session.execute(
            select(
                models.CrossSectionDefinition.width,
                models.CrossSectionDefinition.height,
            )
        )

        error_in_cross_sections = False

        for row in all_cross_sections.all():
            if not is_valid_value(row[0]) or not is_valid_value(row[1]):
                error_in_cross_sections = True
                break  # no need to continue checking; one error is enough to not run the check

        # only run the check if all the cross-section definitions have a parsable width and height
        # otherwise sqlalchemy will throw an exception
        # this is also checked in checks 87 and 88 (CrossSectionFloatListCheck), where it gives an error to the user
        if not error_in_cross_sections:
            return (
                self.to_check(session)
                .filter(self.column == filtered_cross_sections.c.channel_id)
                .all()
            )
        else:
            return []

    def description(self):
        return f"{self.column_name} has both open and closed cross-sections along its length. All cross-sections on a {self.column_name} object must be either open or closed."


class Use0DFlowCheck(BaseCheck):
    """Check that when use_0d_flow in global settings is configured to 1 or to
    2, there is at least one impervious surface or surfaces respectively.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(column=models.GlobalSetting.use_0d_inflow, *args, **kwargs)

    def to_check(self, session):
        """Return a Query object on which this check is applied"""
        return session.query(models.GlobalSetting).filter(
            models.GlobalSetting.use_0d_inflow != 0
        )

    def get_invalid(self, session):
        surface_count = session.query(func.count(models.Surface.id)).scalar()
        impervious_surface_count = session.query(
            func.count(models.ImperviousSurface.id)
        ).scalar()

        invalid_rows = []
        for row in self.to_check(session):
            if (
                row.use_0d_inflow == constants.InflowType.IMPERVIOUS_SURFACE
                and impervious_surface_count == 0
            ):
                invalid_rows.append(row)
            elif (
                row.use_0d_inflow == constants.InflowType.SURFACE and surface_count == 0
            ):
                invalid_rows.append(row)
            else:
                continue
        return invalid_rows

    def description(self):
        return (
            "When %s is used, there should exist at least one "
            "(impervious) surface." % self.column
        )


class ConnectionNodes(BaseCheck):
    """Check that all connection nodes are connected to at least one of the
    following objects:
    - Culvert
    - Channel
    - Pipe
    - Orifice
    - Pumpstation
    - Weir
    """

    def __init__(self, *args, **kwargs):
        super().__init__(column=models.ConnectionNode.id, *args, **kwargs)

    def get_invalid(self, session):
        raise NotImplementedError


class ConnectionNodesLength(BaseCheck):
    """Check that the distance between `start_node` and `end_node` is at least
    `min_distance`. The coords will be transformed into (the first entry) of
    GlobalSettings.epsg_code.
    """

    def __init__(self, start_node, end_node, min_distance: float, *args, **kwargs):
        """

        :param start_node: column name of the start node
        :param end_node: column name of the end node
        :param min_distance: minimum required distance between start and end node
        """
        super().__init__(*args, **kwargs)
        self.start_node = start_node
        self.end_node = end_node
        self.min_distance = min_distance

    def get_invalid(self, session):
        start_node = aliased(models.ConnectionNode)
        end_node = aliased(models.ConnectionNode)
        q = (
            Query(self.column.class_)
            .join(start_node, self.start_node)
            .join(end_node, self.end_node)
            .filter(
                distance(start_node.the_geom, end_node.the_geom) < self.min_distance
            )
        )
        return list(q.with_session(session).all())

    def description(self) -> str:
        return (
            f"The length of {self.table} is "
            f"very short (< {self.min_distance}). A length of at least 1.0 m is recommended."
        )


class ConnectionNodesDistance(BaseCheck):
    """Check that the distance between connection nodes is above a certain
    threshold
    """

    def __init__(
        self, minimum_distance: float, level=CheckLevel.WARNING, *args, **kwargs
    ):
        """
        :param minimum_distance: threshold distance in degrees
        """
        super().__init__(column=models.ConnectionNode.id, level=level, *args, **kwargs)
        self.minimum_distance = minimum_distance

    def get_invalid(self, session: Session) -> List[NamedTuple]:
        """
        The query makes use of the SpatialIndex so we won't have to calculate the
        distance between all connection nodes.
        """
        query = text(
            f"""SELECT *
               FROM v2_connection_nodes AS cn1, v2_connection_nodes AS cn2
               WHERE
                   distance(cn1.the_geom, cn2.the_geom, 1) < :min_distance
                   AND cn1.ROWID != cn2.ROWID
                   AND cn2.ROWID IN (
                     SELECT ROWID
                     FROM SpatialIndex
                     WHERE (
                       f_table_name = "v2_connection_nodes"
                       AND search_frame = Buffer(cn1.the_geom, {self.minimum_distance / 2})));
            """
        )
        results = (
            session.connection()
            .execute(query, {"min_distance": self.minimum_distance})
            .fetchall()
        )

        return results

    def description(self) -> str:
        return (
            f"The connection_node is within {self.minimum_distance} degrees of "
            f"another connection_node."
        )


class ChannelManholeLevelCheck(BaseCheck):
    """Check that the reference_level of a channel is higher than or equal to the bottom_level of a manhole
    connected to the channel as measured at the cross-section closest to the manhole. This check runs if the
    manhole is on the channel's starting node.
    """

    def __init__(
        self,
        level: CheckLevel = CheckLevel.INFO,
        nodes_to_check: Literal["start", "end"] = "start",
        *args,
        **kwargs,
    ):
        """
        :param level: severity of the check, defaults to CheckLevel.INFO. Options are
        in checks.base.CheckLevel
        :param nodes_to_check: whether to check for manholes at the start of the channel
        or at the end of the channel. Options are "start" and "end", defaults to "start"
        """
        if nodes_to_check not in ["start", "end"]:
            raise ValueError("nodes_to_check must be 'start' or 'end'")
        super().__init__(column=models.Channel.id, level=level, *args, **kwargs)
        self.nodes_to_check = nodes_to_check

    def get_invalid(self, session: Session) -> List[NamedTuple]:
        """
        This query does the following:
        channel_with_cs_locations       : left join between cross_sections and channels, to get a table containing
                                          all cross-sections and the channels they lie on
        channels_with_manholes          : join between channel_with_cs_locations and manholes, to get all channels with
                                          a manhole on the channel's start node if self.nodes_to_check == "start", or all
                                          channels with a manhole on the channel's end node if self.nodes_to_check == "start".
        channels_manholes_level_checked : filter the query on invalid entries; that is, entries where the cross-section
                                          reference level is indeed lower than the manhole bottom level. having is used instead
                                          of filter because the query being filtered is a aggregate produced by groupby.
        """
        if self.nodes_to_check == "start":
            func_agg = func.MIN
            connection_node_id_col = models.Channel.connection_node_start_id
        else:
            func_agg = func.MAX
            connection_node_id_col = models.Channel.connection_node_end_id

        channels_with_cs_locations = (
            session.query(
                models.Channel.id,
                models.CrossSectionLocation,
                func_agg(
                    func.Line_Locate_Point(
                        models.Channel.the_geom, models.CrossSectionLocation.the_geom
                    )
                ),
            )
            .join(models.Channel, isouter=True)
            .group_by(models.Channel.id)
        )
        channels_with_manholes = channels_with_cs_locations.join(
            models.Manhole,
            connection_node_id_col == models.Manhole.connection_node_id,
        )
        channels_manholes_level_checked = channels_with_manholes.having(
            models.CrossSectionLocation.reference_level < models.Manhole.bottom_level
        )

        return channels_manholes_level_checked.all()

    def description(self) -> str:
        return (
            f"The v2_manhole.bottom_level at the {self.nodes_to_check} of this v2_channel is higher than the "
            "v2_cross_section_location.reference_level closest to the manhole. This will be "
            "automatically fixed in threedigrid-builder."
        )


class OpenChannelsWithNestedNewton(BaseCheck):
    """Checks whether the model has any closed cross-section in use when the
    NumericalSettings.use_of_nested_newton is turned off.

    See https://github.com/nens/threeditoolbox/issues/522
    """

    def __init__(self, level=CheckLevel.WARNING, *args, **kwargs):
        super().__init__(
            column=models.CrossSectionDefinition.id,
            level=level,
            filters=Query(models.NumericalSettings)
            .filter(models.NumericalSettings.use_of_nested_newton == 0)
            .exists(),
            *args,
            **kwargs,
        )

    def get_invalid(self, session: Session) -> List[NamedTuple]:
        definitions_in_use = self.to_check(session).filter(
            models.CrossSectionDefinition.id.in_(
                Query(models.CrossSectionLocation.definition_id).union_all(
                    Query(models.Pipe.cross_section_definition_id),
                    Query(models.Culvert.cross_section_definition_id),
                    Query(models.Weir.cross_section_definition_id),
                    Query(models.Orifice.cross_section_definition_id),
                )
            ),
        )

        # closed_rectangle, circle, and egg cross-section definitions are always closed:
        closed_definitions = definitions_in_use.filter(
            models.CrossSectionDefinition.shape.in_(
                [
                    constants.CrossSectionShape.CLOSED_RECTANGLE,
                    constants.CrossSectionShape.CIRCLE,
                    constants.CrossSectionShape.EGG,
                ]
            )
        )
        result = list(closed_definitions.with_session(session).all())

        # tabulated cross-section definitions are closed when the last element of 'width'
        # is zero
        tabulated_definitions = definitions_in_use.filter(
            models.CrossSectionDefinition.shape.in_(
                [
                    constants.CrossSectionShape.TABULATED_RECTANGLE,
                    constants.CrossSectionShape.TABULATED_TRAPEZIUM,
                ]
            )
        )
        for definition in tabulated_definitions.with_session(session).all():
            try:
                if float(definition.width.split(" ")[-1]) == 0.0:
                    # Closed channel
                    result.append(definition)
            except Exception:
                # Many things can go wrong, these are caught elsewhere
                pass
        return result

    def description(self) -> str:
        return (
            f"{self.column_name} has a closed cross section definition while "
            f"NumericalSettings.use_of_nested_newton is switched off. "
            f"This gives convergence issues. We recommend setting use_of_nested_newton = 1."
        )


class LinestringLocationCheck(BaseCheck):
    """Check that linestring geometry starts / ends are close to their connection nodes

    This allows for reversing the geometries. threedi-gridbuilder will reverse the geometries if
    that lowers the distance to the connection nodes.
    """

    def __init__(self, *args, **kwargs):
        self.max_distance = kwargs.pop("max_distance")
        super().__init__(*args, **kwargs)

    def get_invalid(self, session: Session) -> List[NamedTuple]:
        start_node = aliased(models.ConnectionNode)
        end_node = aliased(models.ConnectionNode)

        tol = self.max_distance
        start_point = func.ST_PointN(self.column, 1)
        end_point = func.ST_PointN(self.column, func.ST_NPoints(self.column))

        start_ok = distance(start_point, start_node.the_geom) <= tol
        end_ok = distance(end_point, end_node.the_geom) <= tol
        start_ok_if_reversed = distance(end_point, start_node.the_geom) <= tol
        end_ok_if_reversed = distance(start_point, end_node.the_geom) <= tol
        return (
            self.to_check(session)
            .join(start_node, start_node.id == self.table.c.connection_node_start_id)
            .join(end_node, end_node.id == self.table.c.connection_node_end_id)
            .filter(
                ~(start_ok & end_ok),
                ~(start_ok_if_reversed & end_ok_if_reversed),
            )
            .all()
        )

    def description(self) -> str:
        return f"{self.column_name} does not start or end at its connection node (tolerance = {self.max_distance} m)"


class BoundaryCondition1DObjectNumberCheck(BaseCheck):
    """Check that the number of connected objects to 1D boundary connections is 1."""

    def __init__(self, *args, **kwargs):
        super().__init__(
            column=models.BoundaryCondition1D.connection_node_id, *args, **kwargs
        )

    def get_invalid(self, session: Session) -> List[NamedTuple]:
        invalid_ids = []
        for bc in self.to_check(session).all():
            total_objects = 0
            for table in [
                models.Channel,
                models.Pipe,
                models.Culvert,
                models.Orifice,
                models.Weir,
            ]:
                total_objects += (
                    session.query(table)
                    .filter(table.connection_node_start_id == bc.connection_node_id)
                    .count()
                )
                total_objects += (
                    session.query(table)
                    .filter(table.connection_node_end_id == bc.connection_node_id)
                    .count()
                )
            if total_objects != 1:
                invalid_ids.append(bc.id)

        return (
            self.to_check(session)
            .filter(models.BoundaryCondition1D.id.in_(invalid_ids))
            .all()
        )

    def description(self) -> str:
        return "1D boundary condition should be connected to exactly one object."


@dataclass
class IndexMissingRecord:
    id: int
    table_name: str
    column_name: str


class SpatialIndexCheck(BaseCheck):
    """Checks whether a spatial index is present and valid"""

    def get_invalid(self, session: Session) -> List[NamedTuple]:
        result = session.execute(
            func.CheckSpatialIndex(self.column.table.name, self.column.name)
        ).scalar()
        if result == 1:
            return []
        else:
            return [
                IndexMissingRecord(
                    id=1,
                    table_name=self.column.table.name,
                    column_name=self.column.name,
                )
            ]

    def description(self) -> str:
        return f"{self.column_name} has no valid spatial index, which is required for some checks"


class PotentialBreachStartEndCheck(BaseCheck):
    """Check that a potential breach is exactly on or >=1 m from a linestring start/end."""

    def __init__(self, *args, **kwargs):
        self.min_distance = kwargs.pop("min_distance")

        super().__init__(*args, **kwargs)

    def get_invalid(self, session: Session) -> List[NamedTuple]:
        linestring = models.Channel.the_geom
        tol = self.min_distance
        breach_point = func.Line_Locate_Point(
            transform(linestring), transform(func.ST_PointN(self.column, 1))
        )
        dist_1 = breach_point * length(linestring)
        dist_2 = (1 - breach_point) * length(linestring)
        return (
            self.to_check(session)
            .join(models.Channel)
            .filter(((dist_1 > 0) & (dist_1 < tol)) | ((dist_2 > 0) & (dist_2 < tol)))
            .all()
        )

    def description(self) -> str:
        return f"{self.column_name} must be exactly on or >= {self.min_distance} m from a start or end channel vertex"


class PotentialBreachInterdistanceCheck(BaseCheck):
    """Check that a potential breaches are exactly on the same place or >=1 m apart."""

    def __init__(self, *args, **kwargs):
        self.min_distance = kwargs.pop("min_distance")
        assert "filters" not in kwargs

        super().__init__(*args, **kwargs)

    def get_invalid(self, session: Session) -> List[NamedTuple]:
        # this query is hard to get performant; we do a hybrid sql / Python approach

        # First fetch the position of each potential breach per channel
        def get_position(point, linestring):
            breach_point = func.Line_Locate_Point(
                transform(linestring), transform(func.ST_PointN(point, 1))
            )
            return (breach_point * length(linestring)).label("position")

        potential_breaches = sorted(
            session.query(
                self.table, get_position(self.column, models.Channel.the_geom)
            )
            .join(models.Channel)
            .all(),
            key=lambda x: (x.channel_id, x[-1]),
        )

        invalid = []
        prev_channel_id = -9999
        prev_position = -1.0
        for breach in potential_breaches:
            if breach.channel_id != prev_channel_id:
                prev_channel_id, prev_position = breach.channel_id, breach.position
                continue
            if breach.position == prev_position:
                continue
            if (breach.position - prev_position) <= self.min_distance:
                invalid.append(breach)
        return invalid

    def description(self) -> str:
        return f"{self.column_name} must be more than {self.min_distance} m apart (or exactly on the same position)"


class PumpStorageTimestepCheck(BaseCheck):
    """Check that a pumpstation will not empty its storage area within one timestep"""

    def get_invalid(self, session: Session) -> List[NamedTuple]:
        return (
            session.query(models.Pumpstation)
            .join(
                models.ConnectionNode,
                models.Pumpstation.connection_node_start_id == models.ConnectionNode.id,
            )
            .filter(
                (models.ConnectionNode.storage_area != None)
                & (
                    (  # calculate how many seconds the pumpstation takes to empty its storage: (storage * height)/pump capacity
                        (
                            # Arithmetic operations on None return None, so without this
                            # conditional type cast, no invalid results would be returned
                            # even if the storage_area was set to None.
                            models.ConnectionNode.storage_area
                            * (
                                models.Pumpstation.start_level
                                - models.Pumpstation.lower_stop_level
                            )
                        )
                    )
                    / models.Pumpstation.capacity
                    < Query(models.GlobalSetting.sim_time_step)
                    .filter(first_setting_filter)
                    .scalar_subquery()
                )
            )
            .all()
        )

    def description(self) -> str:
        return f"{self.column_name} will empty its storage faster than one timestep, which can cause simulation instabilities"


class ImperviousNodeInflowAreaCheck(BaseCheck):
    """Check that total inflow area per connection node is no larger than 10000 square metres"""

    def __init__(self, *args, **kwargs):
        super().__init__(column=models.ConnectionNode.id, *args, **kwargs)

    def get_invalid(self, session: Session) -> List[NamedTuple]:
        impervious_surfaces = (
            select(models.ImperviousSurfaceMap.connection_node_id)
            .select_from(models.ImperviousSurfaceMap)
            .join(
                models.ImperviousSurface,
                models.ImperviousSurfaceMap.impervious_surface_id
                == models.ImperviousSurface.id,
            )
            .group_by(models.ImperviousSurfaceMap.connection_node_id)
            .having(func.sum(models.ImperviousSurface.area) > 10000)
        ).subquery()

        return (
            session.query(models.ConnectionNode)
            .filter(
                models.ConnectionNode.id == impervious_surfaces.c.connection_node_id
            )
            .all()
        )

    def description(self) -> str:
        return f"{self.column_name} has a an associated inflow area larger than 10000 m2; this might be an error."


class PerviousNodeInflowAreaCheck(BaseCheck):
    """Check that total inflow area per connection node is no larger than 10000 square metres"""

    def __init__(self, *args, **kwargs):
        super().__init__(column=models.ConnectionNode.id, *args, **kwargs)

    def get_invalid(self, session: Session) -> List[NamedTuple]:
        pervious_surfaces = (
            select(models.SurfaceMap.connection_node_id)
            .select_from(models.SurfaceMap)
            .join(
                models.Surface,
                models.SurfaceMap.surface_id == models.Surface.id,
            )
            .group_by(models.SurfaceMap.connection_node_id)
            .having(func.sum(models.Surface.area) > 10000)
        ).subquery()

        return (
            session.query(models.ConnectionNode)
            .filter(models.ConnectionNode.id == pervious_surfaces.c.connection_node_id)
            .all()
        )

    def description(self) -> str:
        return f"{self.column_name} has a an associated inflow area larger than 10000 m2; this might be an error."


class BetaColumnsCheck(BaseCheck):
    """Check that no beta columns were used in the database"""

    def get_invalid(self, session: Session) -> List[NamedTuple]:
        return session.query(self.table).filter(self.column.isnot(None)).all()

    def description(self) -> str:
        return f"{self.column_name} is a beta feature, which is still under development; please do not use it yet."


class BetaValuesCheck(BaseCheck):
    """Check that no beta features were used in the database"""

    def __init__(
        self,
        column,
        values: list = [],
        filters=None,
        level=CheckLevel.ERROR,
        error_code=0,
    ):
        super().__init__(column, filters, level, error_code)
        self.values = values

    def get_invalid(self, session: Session) -> List[NamedTuple]:
        return session.query(self.table).filter(self.column.in_(self.values)).all()

    def description(self) -> str:
        return f"The value you have used for {self.column_name} is still in beta; please do not use it yet."



import pytest
from threedi_schema import models

from threedi_modelchecker.checks.base import CheckLevel
from threedi_modelchecker.checks.factories import (
    generate_enum_checks,
    generate_foreign_key_checks,
    generate_geometry_checks,
    generate_not_null_checks,
    generate_unique_checks,
)


def test_gen_foreign_key_checks():
    foreign_key_checks = generate_foreign_key_checks(models.Manhole.__table__)
    assert len(foreign_key_checks) == 1
    fk_check = foreign_key_checks[0]
    assert models.Manhole.connection_node_id == fk_check.column
    assert models.ConnectionNode.id == fk_check.reference_column


def test_gen_not_unique_checks():
    not_unique_checks = generate_unique_checks(models.Manhole.__table__)
    assert len(not_unique_checks) == 2
    assert models.Manhole.id == not_unique_checks[0].column
    assert models.Manhole.connection_node_id == not_unique_checks[1].column


def test_gen_not_null_checks():
    not_null_checks = generate_not_null_checks(models.Manhole.__table__)
    assert len(not_null_checks) == 3
    not_null_check_columns = [check.column for check in not_null_checks]
    assert models.Manhole.id in not_null_check_columns


def test_gen_geometry_check():
    geometry_checks = generate_geometry_checks(models.ConnectionNode.__table__)

    assert len(geometry_checks) == 2
    geometry_check_columns = [check.column for check in geometry_checks]
    assert models.ConnectionNode.the_geom in geometry_check_columns


def test_gen_enum_checks():
    enum_checks = generate_enum_checks(models.BoundaryConditions2D.__table__)

    assert len(enum_checks) == 1
    assert enum_checks[0].column == models.BoundaryConditions2D.boundary_type


def test_gen_enum_checks_varcharenum():
    enum_checks = generate_enum_checks(models.AggregationSettings.__table__)

    assert len(enum_checks) == 2
    enum_check_columns = [check.column for check in enum_checks]
    assert models.AggregationSettings.aggregation_method in enum_check_columns
    assert models.AggregationSettings.flow_variable in enum_check_columns


@pytest.mark.parametrize(
    "name", ["*.aggregation_method", "v2_aggregation_settings.aggregation_method"]
)
def test_gen_enum_checks_custom_mapping(name):
    enum_checks = generate_enum_checks(
        models.AggregationSettings.__table__,
        custom_level_map={name: "WARNING"},
    )

    assert len(enum_checks) == 2
    checks = {check.column.name: check for check in enum_checks}
    assert checks["aggregation_method"].level == CheckLevel.WARNING
    assert checks["flow_variable"].level == CheckLevel.ERROR



import click
from threedi_schema import ThreediDatabase
from threedi_schema.domain.models import DECLARED_MODELS

from threedi_modelchecker import exporters
from threedi_modelchecker.checks.base import CheckLevel
from threedi_modelchecker.config import Config
from threedi_modelchecker.model_checks import ThreediModelChecker


@click.group()
def cli():
    pass


@cli.command()
@click.option("-f", "--file", help="Write errors to file, instead of stdout")
@click.option(
    "-l",
    "--level",
    type=click.Choice([x.name for x in CheckLevel], case_sensitive=False),
    default="ERROR",
    help="Minimum check level.",
)
@click.option(
    "-s",
    "--sqlite",
    type=click.Path(exists=True, readable=True),
    help="Path to an sqlite (spatialite) file",
    required=True,
)
@click.option(
    "--allow-beta",
    is_flag=True,
    default=False,
    help="Don't check whether beta features were used in the database.",
)
def check(sqlite, file, level, allow_beta):
    """Checks the threedi-model for errors / warnings / info messages"""
    db = ThreediDatabase(sqlite, echo=False)
    """Checks the threedi model schematisation for errors."""
    level = level.upper()
    if level == "ERROR":
        msg = "errors"
    elif level == "WARNING":
        msg = "errors or warnings"
    else:
        msg = "errors, warnings or info messages"
    click.echo("Parsing schematisation for any %s" % msg)
    if file:
        click.echo("Model errors will be written to %s" % file)

    mc = ThreediModelChecker(threedi_db=db, allow_beta_features=allow_beta)
    model_errors = mc.errors(level=level)

    if file:
        exporters.export_to_file(model_errors, file)
    else:
        exporters.print_errors(model_errors)

    click.echo("Finished processing model")


@cli.command()
@click.option("-f", "--file", help="Write output to file, instead of stdout")
@click.option(
    "-ft",
    "--format",
    type=click.Choice(["rst", "csv"], case_sensitive=False),
    default="rst",
    help="Export format for checks table",
)
def export_checks(file, format):
    """Export formatted checks summary to insert in documentation or use elsewhere"""
    checks = Config(models=DECLARED_MODELS).checks

    if format.lower() == "rst":
        table = exporters.generate_rst_table(checks=checks)
    elif format.lower() == "csv":
        table = exporters.generate_csv_table(checks=checks)
    if file:
        with open(file, "w") as f:
            f.write(table)
    else:
        click.echo(table)


if __name__ == "__main__":
    check()



from .raster_interface import RasterInterface

try:
    from osgeo import gdal, osr

    gdal.UseExceptions()
    osr.UseExceptions()
except ImportError:
    gdal = osr = None


class GDALRasterInterface(RasterInterface):
    def __init__(self, path):
        if gdal is None:
            raise ImportError("This raster check requires GDAL")
        super().__init__(path)

    @staticmethod
    def available():
        return gdal is not None

    def _open(self):
        try:
            self._dataset = gdal.Open(self.path, gdal.GA_ReadOnly)
        except RuntimeError:
            self._dataset = None

    def _close(self):
        self._dataset = None

    @property
    def _spatial_reference(self):
        dataset = self._dataset
        projection = None if dataset == None else dataset.GetProjection()
        if projection:
            return osr.SpatialReference(projection)

    @property
    def is_valid_geotiff(self):
        return (
            self._dataset is not None and self._dataset.GetDriver().ShortName == "GTiff"
        )

    @property
    def band_count(self):
        return self._dataset.RasterCount

    @property
    def has_projection(self) -> bool:
        return self._spatial_reference is not None

    @property
    def is_geographic(self) -> bool:
        sr = self._spatial_reference
        return bool(sr.IsGeographic())

    @property
    def epsg_code(self):
        code = self._spatial_reference.GetAuthorityCode("PROJCS")
        return int(code) if code is not None else None

    @property
    def pixel_size(self):
        gt = self._dataset.GetGeoTransform()
        if gt is None:
            return None, None
        else:
            return abs(gt[1]), abs(gt[5])

    @property
    def min_max(self):
        if self.band_count == 0:
            return None, None
        # usage of approx_ok=False bypasses statistics cache and forces
        # all pixels to be read
        # see: https://gdal.org/doxygen/classGDALRasterBand.html#ac7761bab7cf3b8445ed963e4aa85e715
        try:
            return self._dataset.GetRasterBand(1).ComputeRasterMinMax(False)
        except RuntimeError as e:
            if "no valid pixels found" in str(e):
                raise self.NoData()
            else:
                raise e



from unittest import mock

from threedi_schema import models

from threedi_modelchecker.checks.raster import (
    BaseRasterCheck,
    GDALAvailableCheck,
    LocalContext,
    RasterExistsCheck,
    RasterGridSizeCheck,
    RasterHasMatchingEPSGCheck,
    RasterHasOneBandCheck,
    RasterHasProjectionCheck,
    RasterIsProjectedCheck,
    RasterIsValidCheck,
    RasterRangeCheck,
    RasterSquareCellsCheck,
    ServerContext,
)
from threedi_modelchecker.interfaces.raster_interface_gdal import GDALRasterInterface
from threedi_modelchecker.interfaces.raster_interface_rasterio import (
    RasterIORasterInterface,
)

from . import factories

try:
    import numpy as np
    from osgeo import gdal, osr
except ImportError:
    gdal = osr = np = None

import pytest


@pytest.fixture
def mocked_check():
    with mock.patch.object(BaseRasterCheck, "is_valid", return_value=True):
        yield BaseRasterCheck(column=models.GlobalSetting.dem_file)


@pytest.fixture
def context_local(tmp_path):
    return LocalContext(base_path=tmp_path)


@pytest.fixture
def context_server():
    return ServerContext(available_rasters={})


@pytest.fixture
def session_local(session, context_local):
    session.model_checker_context = context_local
    return session


@pytest.fixture
def session_server(session, context_server):
    session.model_checker_context = context_server
    return session


def create_geotiff(
    path, epsg=28992, width=3, height=2, bands=1, dx=0.5, dy=0.5, value=None
):
    ds = gdal.GetDriverByName("GTiff").Create(
        str(path), width, height, bands, gdal.GDT_Byte
    )
    if epsg is not None:
        if isinstance(epsg, int):
            wkt = osr.GetUserInputAsWKT(f"EPSG:{epsg}")
        else:
            wkt = epsg
        ds.SetProjection(wkt)
    ds.SetGeoTransform((155000.0, dx, 0, 463000.0, 0, -dy))
    band = ds.GetRasterBand(1)
    band.SetNoDataValue(255)
    if value is None:
        data = np.arange(height * width).reshape(height, width)
    else:
        data = np.full((height, width), fill_value=value, dtype=int)
    band.WriteArray(data)
    ds.FlushCache()
    return str(path)


@pytest.fixture
def valid_geotiff(tmp_path):
    return create_geotiff(tmp_path / "raster.tiff")


@pytest.fixture
def invalid_geotiff(tmp_path):
    path = tmp_path / "raster.tiff"
    path.touch()
    return str(path)


def test_base_to_check(session):
    factories.GlobalSettingsFactory(dem_file="somefile")
    check = BaseRasterCheck(column=models.GlobalSetting.dem_file)
    assert check.to_check(session).count() == 1


def test_base_to_check_ignores_empty(session):
    factories.GlobalSettingsFactory(dem_file="")
    check = BaseRasterCheck(column=models.GlobalSetting.dem_file)
    assert check.to_check(session).count() == 0


def test_base_to_check_ignores_none(session):
    factories.GlobalSettingsFactory(dem_file=None)
    check = BaseRasterCheck(column=models.GlobalSetting.dem_file)
    assert check.to_check(session).count() == 0


def test_base_get_invalid_local(mocked_check, session_local, invalid_geotiff):
    factories.GlobalSettingsFactory(dem_file="raster.tiff")
    assert mocked_check.get_invalid(session_local) == []
    mocked_check.is_valid.assert_called_once_with(
        invalid_geotiff, session_local.model_checker_context.raster_interface
    )


def test_base_get_invalid_local_no_file(mocked_check, session_local):
    factories.GlobalSettingsFactory(dem_file="somefile")
    assert mocked_check.get_invalid(session_local) == []
    assert not mocked_check.is_valid.called


def test_base_get_invalid_server(mocked_check, context_server, session_server):
    factories.GlobalSettingsFactory(dem_file="somefile")
    context_server.available_rasters = {"dem_file": "http://tempurl"}
    assert mocked_check.get_invalid(session_server) == []
    mocked_check.is_valid.assert_called_once_with(
        "http://tempurl", session_server.model_checker_context.raster_interface
    )


def test_base_get_invalid_server_no_file(mocked_check, context_server, session_server):
    factories.GlobalSettingsFactory(dem_file="somefile")
    context_server.available_rasters = {"other": "http://tempurl"}
    assert mocked_check.get_invalid(session_server) == []
    assert not mocked_check.is_valid.called


def test_base_get_invalid_server_available_set(
    mocked_check, context_server, session_server
):
    factories.GlobalSettingsFactory(dem_file="somefile")
    context_server.available_rasters = {"dem_file"}
    assert mocked_check.get_invalid(session_server) == []
    assert not mocked_check.is_valid.called


def test_base_no_gdal(mocked_check, session_local):
    with mock.patch.object(
        session_local.model_checker_context.raster_interface,
        "available",
        return_value=False,
    ):
        assert mocked_check.get_invalid(session_local) == []
        assert not mocked_check.is_valid.called


def test_exists_local_ok(session_local, invalid_geotiff):
    factories.GlobalSettingsFactory(dem_file="raster.tiff")
    check = RasterExistsCheck(column=models.GlobalSetting.dem_file)
    assert check.get_invalid(session_local) == []


def test_exists_local_err(session_local):
    factories.GlobalSettingsFactory(dem_file="raster.tiff")
    check = RasterExistsCheck(column=models.GlobalSetting.dem_file)
    assert len(check.get_invalid(session_local)) == 1


@pytest.mark.parametrize(
    "available_rasters", [{"dem_file": "http://tempurl"}, {"dem_file"}]
)
def test_exists_server_ok(session_server, context_server, available_rasters):
    factories.GlobalSettingsFactory(dem_file="raster.tiff")
    check = RasterExistsCheck(column=models.GlobalSetting.dem_file)
    context_server.available_rasters = available_rasters
    assert check.get_invalid(session_server) == []


@pytest.mark.parametrize("available_rasters", [{"other": "http://tempurl"}, {"other"}])
def test_exists_server_err(session_server, context_server, available_rasters):
    factories.GlobalSettingsFactory(dem_file="raster.tiff")
    check = RasterExistsCheck(column=models.GlobalSetting.dem_file)
    context_server.available_rasters = available_rasters
    assert len(check.get_invalid(session_server)) == 1


@pytest.mark.parametrize(
    "interface_cls", [GDALRasterInterface, RasterIORasterInterface]
)
def test_valid_ok(valid_geotiff, interface_cls):
    check = RasterIsValidCheck(column=models.GlobalSetting.dem_file)
    assert check.is_valid(valid_geotiff, interface_cls)


@pytest.mark.parametrize(
    "interface_cls", [GDALRasterInterface, RasterIORasterInterface]
)
def test_valid_err(invalid_geotiff, interface_cls):
    check = RasterIsValidCheck(column=models.GlobalSetting.dem_file)
    assert not check.is_valid(invalid_geotiff, interface_cls)


@pytest.mark.parametrize(
    "interface_cls", [GDALRasterInterface, RasterIORasterInterface]
)
def test_one_band_ok(valid_geotiff, interface_cls):
    check = RasterHasOneBandCheck(column=models.GlobalSetting.dem_file)
    assert check.is_valid(valid_geotiff, interface_cls)


@pytest.mark.parametrize(
    "interface_cls", [GDALRasterInterface, RasterIORasterInterface]
)
def test_one_band_err(tmp_path, interface_cls):
    path = create_geotiff(tmp_path / "raster.tiff", bands=2)
    check = RasterHasOneBandCheck(column=models.GlobalSetting.dem_file)
    assert not check.is_valid(path, interface_cls)


@pytest.mark.parametrize(
    "interface_cls", [GDALRasterInterface, RasterIORasterInterface]
)
def test_has_projection_ok(valid_geotiff, interface_cls):
    check = RasterHasProjectionCheck(column=models.GlobalSetting.dem_file)
    assert check.is_valid(valid_geotiff, interface_cls)


@pytest.mark.parametrize(
    "interface_cls", [GDALRasterInterface, RasterIORasterInterface]
)
def test_has_projection_err(tmp_path, interface_cls):
    path = create_geotiff(tmp_path / "raster.tiff", epsg=None)
    check = RasterHasProjectionCheck(column=models.GlobalSetting.dem_file)
    assert not check.is_valid(path, interface_cls)


NULL_EPSG_CODE = (
    'PROJCS["unnamed",GEOGCS["GRS 1980(IUGG, 1980)"'
    + ',DATUM["unknown",SPHEROID["GRS80",6378137,298.257222101]],'
    + 'PRIMEM["Greenwich",0],UNIT["degree",0.0174532925199433]],'
    + 'PROJECTION["Albers_Conic_Equal_Area"],PARAMETER["standard_parallel_1",-18],'
    + 'PARAMETER["standard_parallel_2",-36],PARAMETER["latitude_of_center",-33.264],'
    + 'PARAMETER["longitude_of_center",150.874],PARAMETER["false_easting",0],'
    + 'PARAMETER["false_northing",0],UNIT["kilometre",1000]]'
)


@pytest.mark.parametrize(
    "interface_cls", [GDALRasterInterface, RasterIORasterInterface]
)
@pytest.mark.parametrize(
    "raster_epsg, sqlite_epsg, validity",
    [
        (28992, 28992, True),
        (28992, 27700, False),
        (NULL_EPSG_CODE, 28992, False),
        (27700, None, False),
    ],
)
def test_has_epsg(tmp_path, interface_cls, raster_epsg, sqlite_epsg, validity):
    path = create_geotiff(tmp_path / "raster.tiff", epsg=raster_epsg)
    check = RasterHasMatchingEPSGCheck(column=models.GlobalSetting.dem_file)
    check.epsg_code = sqlite_epsg
    assert check.is_valid(path, interface_cls) == validity


@pytest.mark.parametrize(
    "interface_cls", [GDALRasterInterface, RasterIORasterInterface]
)
def test_is_projected_ok(valid_geotiff, interface_cls):
    check = RasterIsProjectedCheck(column=models.GlobalSetting.dem_file)
    assert check.is_valid(valid_geotiff, interface_cls)


@pytest.mark.parametrize(
    "interface_cls", [GDALRasterInterface, RasterIORasterInterface]
)
def test_is_projected_err(tmp_path, interface_cls):
    path = create_geotiff(tmp_path / "raster.tiff", epsg=4326)
    check = RasterIsProjectedCheck(column=models.GlobalSetting.dem_file)
    assert not check.is_valid(path, interface_cls)


@pytest.mark.parametrize(
    "interface_cls", [GDALRasterInterface, RasterIORasterInterface]
)
def test_is_projected_no_projection(tmp_path, interface_cls):
    path = create_geotiff(tmp_path / "raster.tiff", epsg=None)
    check = RasterIsProjectedCheck(column=models.GlobalSetting.dem_file)
    assert check.is_valid(path, interface_cls)


@pytest.mark.parametrize(
    "interface_cls", [GDALRasterInterface, RasterIORasterInterface]
)
def test_square_cells_ok(valid_geotiff, interface_cls):
    check = RasterSquareCellsCheck(column=models.GlobalSetting.dem_file)
    assert check.is_valid(valid_geotiff, interface_cls)


@pytest.mark.parametrize(
    "interface_cls", [GDALRasterInterface, RasterIORasterInterface]
)
def test_square_cells_err(tmp_path, interface_cls):
    path = create_geotiff(tmp_path / "raster.tiff", dx=0.5, dy=1.0)
    check = RasterSquareCellsCheck(column=models.GlobalSetting.dem_file)
    assert not check.is_valid(path, interface_cls)


@pytest.mark.parametrize(
    "interface_cls", [GDALRasterInterface, RasterIORasterInterface]
)
def test_square_cells_rounding(tmp_path, interface_cls):
    path = create_geotiff(tmp_path / "raster.tiff", dx=0.5, dy=0.5001)
    check = RasterSquareCellsCheck(decimals=3, column=models.GlobalSetting.dem_file)
    assert check.is_valid(path, interface_cls)
    check = RasterSquareCellsCheck(decimals=4, column=models.GlobalSetting.dem_file)
    assert not check.is_valid(path, interface_cls)


@pytest.mark.parametrize(
    "interface_cls", [GDALRasterInterface, RasterIORasterInterface]
)
@pytest.mark.parametrize(
    "raster_pixel_size, sqlite_grid_space, validity",
    [
        (2, 7, False),
        (2, 4, True),
        (2, 3, False),
        (2, 0, False),
        (2, -4, False),
    ],
)
def test_raster_grid_size(
    tmp_path, interface_cls, raster_pixel_size, sqlite_grid_space, validity
):
    path = create_geotiff(
        tmp_path / "raster.tiff", dx=raster_pixel_size, dy=raster_pixel_size
    )
    check = RasterGridSizeCheck(column=models.GlobalSetting.dem_file)
    check.grid_space = sqlite_grid_space
    assert check.is_valid(path, interface_cls) == validity


@pytest.mark.parametrize(
    "interface_cls", [GDALRasterInterface, RasterIORasterInterface]
)
def test_raster_range_ok(valid_geotiff, interface_cls):
    check = RasterRangeCheck(
        column=models.GlobalSetting.dem_file, min_value=0, max_value=5
    )
    assert check.is_valid(valid_geotiff, interface_cls)


@pytest.mark.parametrize(
    "interface_cls", [GDALRasterInterface, RasterIORasterInterface]
)
@pytest.mark.parametrize(
    "kwargs,msg",
    [
        ({"min_value": 1}, "{} has values <1 or is empty"),
        ({"max_value": 4}, "{} has values >4 or is empty"),
        ({"min_value": 0, "left_inclusive": False}, "{} has values <=0 or is empty"),
        ({"max_value": 5, "right_inclusive": False}, "{} has values >=5 or is empty"),
        ({"min_value": 1, "max_value": 6}, "{} has values <1 and/or >6 or is empty"),
    ],
)
def test_raster_range_err(valid_geotiff, kwargs, msg, interface_cls):
    check = RasterRangeCheck(column=models.GlobalSetting.dem_file, **kwargs)
    assert not check.is_valid(valid_geotiff, interface_cls)
    assert check.description() == msg.format("v2_global_settings.dem_file")


@pytest.mark.parametrize(
    "interface_cls", [GDALRasterInterface, RasterIORasterInterface]
)
def test_raster_range_no_data(tmp_path, interface_cls):
    path = tmp_path / "raster.tiff"
    create_geotiff(path, value=255)
    check = RasterRangeCheck(column=models.GlobalSetting.dem_file, min_value=0)
    assert not check.is_valid(path, interface_cls)


@pytest.mark.parametrize(
    "interface_cls", [GDALRasterInterface, RasterIORasterInterface]
)
@pytest.mark.parametrize(
    "check",
    [
        RasterHasOneBandCheck(column=models.GlobalSetting.dem_file),
        RasterHasProjectionCheck(column=models.GlobalSetting.dem_file),
        RasterHasMatchingEPSGCheck(column=models.GlobalSetting.dem_file),
        RasterIsProjectedCheck(column=models.GlobalSetting.dem_file),
        RasterSquareCellsCheck(column=models.GlobalSetting.dem_file),
        RasterGridSizeCheck(column=models.GlobalSetting.dem_file),
        RasterRangeCheck(column=models.GlobalSetting.dem_file, min_value=0),
    ],
)
def test_raster_check_invalid_file(check, invalid_geotiff, interface_cls):
    assert check.is_valid(invalid_geotiff, interface_cls)


def test_gdal_check_ok(session_local):
    check = GDALAvailableCheck(column=models.GlobalSetting.dem_file)
    assert not check.get_invalid(session_local)


def test_gdal_check_err(session_local):
    with mock.patch.object(
        session_local.model_checker_context.raster_interface,
        "available",
        return_value=False,
    ):
        check = GDALAvailableCheck(column=models.GlobalSetting.dem_file)
        assert check.get_invalid(session_local)



import pytest
from threedi_schema import constants, models

from threedi_modelchecker.checks.cross_section_definitions import (
    CrossSectionEqualElementsCheck,
    CrossSectionExpectEmptyCheck,
    CrossSectionFirstElementNonZeroCheck,
    CrossSectionFloatCheck,
    CrossSectionFloatListCheck,
    CrossSectionGreaterZeroCheck,
    CrossSectionIncreasingCheck,
    CrossSectionNullCheck,
    CrossSectionYZCoordinateCountCheck,
    CrossSectionYZHeightCheck,
    CrossSectionYZIncreasingWidthIfOpenCheck,
)

from . import factories


def test_in_use(session):
    # should only check records in use
    definition = factories.CrossSectionDefinitionFactory(
        width=None, shape=constants.CrossSectionShape.CIRCLE
    )

    check = CrossSectionNullCheck(column=models.CrossSectionDefinition.width)
    invalid_rows = check.get_invalid(session)
    assert len(invalid_rows) == 0

    factories.CrossSectionLocationFactory(definition=definition)
    invalid_rows = check.get_invalid(session)
    assert len(invalid_rows) == 1


def test_filter_shapes(session):
    # should only check records of given types
    definition = factories.CrossSectionDefinitionFactory(
        width=None, shape=constants.CrossSectionShape.CIRCLE
    )
    factories.CrossSectionLocationFactory(definition=definition)

    check = CrossSectionNullCheck(
        column=models.CrossSectionDefinition.width,
        shapes=[constants.CrossSectionShape.RECTANGLE],
    )
    invalid_rows = check.get_invalid(session)
    assert len(invalid_rows) == 0

    check = CrossSectionNullCheck(
        column=models.CrossSectionDefinition.width,
        shapes=[constants.CrossSectionShape.CIRCLE],
    )
    invalid_rows = check.get_invalid(session)
    assert len(invalid_rows) == 1


@pytest.mark.parametrize("width", [None, ""])
def test_check_null_invalid(session, width):
    definition = factories.CrossSectionDefinitionFactory(width=width)
    factories.CrossSectionLocationFactory(definition=definition)
    check = CrossSectionNullCheck(column=models.CrossSectionDefinition.width)
    invalid_rows = check.get_invalid(session)
    assert len(invalid_rows) == 1

    check = CrossSectionExpectEmptyCheck(column=models.CrossSectionDefinition.width)
    invalid_rows = check.get_invalid(session)
    assert len(invalid_rows) == 0


@pytest.mark.parametrize("width", [" "])
def test_check_null_valid(session, width):
    definition = factories.CrossSectionDefinitionFactory(width=width)
    factories.CrossSectionLocationFactory(definition=definition)
    check = CrossSectionNullCheck(column=models.CrossSectionDefinition.width)
    invalid_rows = check.get_invalid(session)
    assert len(invalid_rows) == 0

    check = CrossSectionExpectEmptyCheck(column=models.CrossSectionDefinition.width)
    invalid_rows = check.get_invalid(session)
    assert len(invalid_rows) == 1


@pytest.mark.parametrize("width", [" ", "foo", "0,1", "1e-2e8", "-0.1"])
def test_check_float_invalid(session, width):
    definition = factories.CrossSectionDefinitionFactory(width=width)
    factories.CrossSectionLocationFactory(definition=definition)
    check = CrossSectionFloatCheck(column=models.CrossSectionDefinition.width)
    invalid_rows = check.get_invalid(session)
    assert len(invalid_rows) == 1


@pytest.mark.parametrize(
    "width", [None, "", "2", "0.1", ".2", "7.", "1e-5", "1E+2", " 0.1"]
)
def test_check_float_valid(session, width):
    definition = factories.CrossSectionDefinitionFactory(width=width)
    factories.CrossSectionLocationFactory(definition=definition)
    check = CrossSectionFloatCheck(column=models.CrossSectionDefinition.width)
    invalid_rows = check.get_invalid(session)
    assert len(invalid_rows) == 0


@pytest.mark.parametrize("width", ["0", " 0", "0.0", "-0", "-1.2"])
def test_check_greater_zero_invalid(session, width):
    definition = factories.CrossSectionDefinitionFactory(width=width)
    factories.CrossSectionLocationFactory(definition=definition)
    check = CrossSectionGreaterZeroCheck(column=models.CrossSectionDefinition.width)
    invalid_rows = check.get_invalid(session)
    assert len(invalid_rows) == 1


@pytest.mark.parametrize("width", [None, "", "foo", "0.1", "1e-2"])
def test_check_greater_zero_valid(session, width):
    definition = factories.CrossSectionDefinitionFactory(width=width)
    factories.CrossSectionLocationFactory(definition=definition)
    check = CrossSectionGreaterZeroCheck(column=models.CrossSectionDefinition.width)
    invalid_rows = check.get_invalid(session)
    assert len(invalid_rows) == 0


@pytest.mark.parametrize("width", [" ", "0,1,2", "3;5;7", "foo"])
def test_check_float_list_invalid(session, width):
    definition = factories.CrossSectionDefinitionFactory(width=width)
    factories.CrossSectionLocationFactory(definition=definition)
    check = CrossSectionFloatListCheck(column=models.CrossSectionDefinition.width)
    invalid_rows = check.get_invalid(session)
    assert len(invalid_rows) == 1


@pytest.mark.parametrize("width", [None, "", "0", "0.1", "0 1 2", "-.2 5.72 9. 1e2"])
def test_check_float_list_valid(session, width):
    definition = factories.CrossSectionDefinitionFactory(width=width)
    factories.CrossSectionLocationFactory(definition=definition)
    check = CrossSectionFloatListCheck(column=models.CrossSectionDefinition.width)
    invalid_rows = check.get_invalid(session)
    assert len(invalid_rows) == 0


@pytest.mark.parametrize("width", ["0", "0 1"])
def test_check_equal_elements_invalid(session, width):
    definition = factories.CrossSectionDefinitionFactory(
        width=width,
        height="0 2 5",
    )
    factories.CrossSectionLocationFactory(definition=definition)
    check = CrossSectionEqualElementsCheck()
    invalid_rows = check.get_invalid(session)
    assert len(invalid_rows) == 1


@pytest.mark.parametrize("width", [None, "", "3;5;7", "1 2 3"])
def test_check_equal_elements_valid(session, width):
    definition = factories.CrossSectionDefinitionFactory(
        width=width,
        height="0 2 5",
    )
    factories.CrossSectionLocationFactory(definition=definition)
    check = CrossSectionEqualElementsCheck()
    invalid_rows = check.get_invalid(session)
    assert len(invalid_rows) == 0


@pytest.mark.parametrize("width", ["2 1 4"])
def test_increasing_elements_invalid(session, width):
    definition = factories.CrossSectionDefinitionFactory(width=width)
    factories.CrossSectionLocationFactory(definition=definition)
    check = CrossSectionIncreasingCheck(column=models.CrossSectionDefinition.width)
    invalid_rows = check.get_invalid(session)
    assert len(invalid_rows) == 1


@pytest.mark.parametrize("width", [None, "", "3;5;7", "1 2 3"])
def test_increasing_elements_valid(session, width):
    definition = factories.CrossSectionDefinitionFactory(width=width)
    factories.CrossSectionLocationFactory(definition=definition)
    check = CrossSectionIncreasingCheck(column=models.CrossSectionDefinition.width)
    invalid_rows = check.get_invalid(session)
    assert len(invalid_rows) == 0


@pytest.mark.parametrize("width", ["0 1 4"])
def test_first_nonzero_invalid(session, width):
    definition = factories.CrossSectionDefinitionFactory(width=width)
    factories.CrossSectionLocationFactory(definition=definition)
    check = CrossSectionFirstElementNonZeroCheck(
        column=models.CrossSectionDefinition.width
    )
    invalid_rows = check.get_invalid(session)
    assert len(invalid_rows) == 1


@pytest.mark.parametrize("width", [None, "", "3;5;7", "1 2 3"])
def test_first_nonzero_valid(session, width):
    definition = factories.CrossSectionDefinitionFactory(width=width)
    factories.CrossSectionLocationFactory(definition=definition)
    check = CrossSectionFirstElementNonZeroCheck(
        column=models.CrossSectionDefinition.width
    )
    invalid_rows = check.get_invalid(session)
    assert len(invalid_rows) == 0


@pytest.mark.parametrize("height", ["0 1 2", "0 1 1", "1 0 1", "foo", None, "0"])
def test_check_yz_height_valid(session, height):
    definition = factories.CrossSectionDefinitionFactory(
        width="1 2 3",
        height=height,
    )
    factories.CrossSectionLocationFactory(definition=definition)
    check = CrossSectionYZHeightCheck(column=models.CrossSectionDefinition.height)
    invalid_rows = check.get_invalid(session)
    assert len(invalid_rows) == 0


@pytest.mark.parametrize("height", ["1 2 3", "0 -1 1"])
def test_check_yz_height_invalid(session, height):
    definition = factories.CrossSectionDefinitionFactory(
        width="1 2 3",
        height=height,
    )
    factories.CrossSectionLocationFactory(definition=definition)
    check = CrossSectionYZHeightCheck(column=models.CrossSectionDefinition.height)
    invalid_rows = check.get_invalid(session)
    assert len(invalid_rows) == 1


@pytest.mark.parametrize(
    "width,height",
    [
        # ("0 0.5 1 1.5", "0.5 0 0 0.5"),
        ("0 0.5", "0.5 0"),
        ("0 0.5 0", "0.5 0 0.5"),
    ],
)
def test_check_yz_coord_count_invalid(session, width, height):
    definition = factories.CrossSectionDefinitionFactory(
        width=width,
        height=height,
    )
    factories.CrossSectionLocationFactory(definition=definition)
    check = CrossSectionYZCoordinateCountCheck()
    invalid_rows = check.get_invalid(session)
    assert len(invalid_rows) == 1


@pytest.mark.parametrize(
    "width,height",
    [
        ("0 0.5 1 1.5", "0.5 0 0 0.5"),
        ("0.5 0 0.5 1.5 1.5 0.5", "0 1 2 2 0 0"),
        ("foo", ""),
        ("0 0.5", "0.5"),
        ("0 0.5 1 1.5 2.0", "0.5 0 0 0.5"),
    ],
)
def test_check_yz_coord_count_valid(session, width, height):
    definition = factories.CrossSectionDefinitionFactory(
        width=width,
        height=height,
    )
    factories.CrossSectionLocationFactory(definition=definition)
    check = CrossSectionYZCoordinateCountCheck()
    invalid_rows = check.get_invalid(session)
    assert len(invalid_rows) == 0


@pytest.mark.parametrize(
    "width,height",
    [
        ("0 0.5 1 1", "0.5 0 0 0.5"),
        ("0.5 0 0.5 1.5 1.5 0.5", "0 1 2 2 0 1"),
    ],
)
def test_check_yz_increasing_if_open_invalid(session, width, height):
    definition = factories.CrossSectionDefinitionFactory(
        width=width,
        height=height,
    )
    factories.CrossSectionLocationFactory(definition=definition)
    check = CrossSectionYZIncreasingWidthIfOpenCheck()
    invalid_rows = check.get_invalid(session)
    assert len(invalid_rows) == 1


@pytest.mark.parametrize(
    "width,height",
    [
        ("0 0.5 1 1.5", "0.5 0 0 0.5"),
        ("0.5 0 0.5 1.5 1.5 0.5", "0 1 2 2 0 0"),
        ("foo", ""),
        ("0 0.5", "0.5"),
        ("0 0.5 1 1.5 2.0", "0.5 0 0 0.5"),
    ],
)
def test_check_yz_increasing_if_open_valid(session, width, height):
    definition = factories.CrossSectionDefinitionFactory(
        width=width,
        height=height,
    )
    factories.CrossSectionLocationFactory(definition=definition)
    check = CrossSectionYZIncreasingWidthIfOpenCheck()
    invalid_rows = check.get_invalid(session)
    assert len(invalid_rows) == 0
